(* author: Rene Thiemann *)

(* this theory provides an alternative code setup for characters and strings, 
   where in particular characters are represented by integers and not by 8 booleans,
   so that execution is faster *)

theory Code_Char_Setup
  imports 
    Main
    "HOL-Library.Char_ord" 
    "HOL-Library.List_Lexorder" 
begin

(* representative type for characters *)

definition bound_ci :: "integer \<Rightarrow> bool" where [simp]: "bound_ci x = (0 \<le> x \<and> x < 256)" 

typedef char_integer = "{x :: integer . bound_ci x}" by auto

setup_lifting type_definition_char_integer

lift_definition char_of_ci :: "char_integer \<Rightarrow> char" is char_of_integer .
lift_definition integer_of_ci :: "char_integer \<Rightarrow> integer" is "\<lambda> x. x" .
lift_definition ascii_of_ci :: "char_integer \<Rightarrow> char_integer" is "\<lambda> x. if x < 128 then x else x - 128" by simp

lift_definition (code_dt) char_integer_of_integer :: "integer \<Rightarrow> char_integer option" 
  is "\<lambda> x. if 0 \<le> x \<and> x < 256 then Some x else None" 
  by auto

(* now perform setup of alternative char-representation *)

code_datatype char_of_ci

declare [[code drop: 
      integer_of_char 
      char_of_integer
      String.ascii_of 
      String.asciis_of_literal 
      String.literal_of_asciis 
      of_char]]

declare 
  equal_char_def[code del]
  less_char_def[code del] 
  less_eq_char_def[code del]

lemma integer_of_char_code[code]: "integer_of_char (char_of_ci x) = integer_of_ci x" 
  apply (transfer, auto simp: integer_of_char_def char_of_integer_def)
  by (simp add: unique_euclidean_semiring_numeral_class.mod_less)

lemma char_of_integer_code[code]: "char_of_integer x = (case char_integer_of_integer x of None \<Rightarrow> 
    Code.abort (STR ''invalid character code'') (\<lambda> _. char_of_integer x)
  | Some y \<Rightarrow> char_of_ci y)"
  by (transfer, auto simp: char_of_integer_def)

lemma digit7_char_of[simp]: "0 \<le> (c :: integer) \<Longrightarrow> c < 128 \<Longrightarrow> \<not> digit7 (char_of c)" 
  by (simp add: char_of_def bit_iff_odd unique_euclidean_semiring_numeral_class.div_less)

lemma bit_add: fixes x :: integer 
  assumes "i < n" "0 \<le> x" "x < 2^n" shows "bit (x + 2^n) i = bit x i" 
proof -
  define j where "j = n - i" 
  from assms(1) have n: "n = i + j" and j: "j > 0" unfolding j_def by auto
  show ?thesis unfolding n power_add bit_iff_odd using j by simp
qed

lemma ascii_of_code[code]: "String.ascii_of (char_of_ci c) = char_of_ci (ascii_of_ci c)" 
proof (transfer, goal_cases)
  case (1 c)
  show ?case
  proof (cases "c < 128")
    case True
    thus ?thesis using 1 
      by (simp add: char_of_integer_def String.ascii_of_def char_of_def bit_iff_odd unique_euclidean_semiring_numeral_class.div_less)
  next
    case False
    define d where "d = c - 128"
    from False have "0 \<le> d" "d < 128" "c = d + 128" using 1 unfolding d_def by auto
    thus ?thesis apply (simp add: String.ascii_of_def char_of_integer_def char_of_def, intro conjI)
      subgoal using bit_add[of 1 7 d] by simp
      subgoal using bit_add[of 2 7 d] by simp
      subgoal using bit_add[of 3 7 d] by simp
      subgoal using bit_add[of 4 7 d] by simp
      subgoal using bit_add[of 5 7 d] by simp
      subgoal using bit_add[of 6 7 d] by simp
      subgoal unfolding bit_iff_odd by (simp add: unique_euclidean_semiring_numeral_class.div_less)
      done
  qed
qed

lemma integer_of_char_of_nat: "integer_of_char c = integer_of_nat (of_char c :: nat)" 
proof -
  define ds where "ds = [digit0 c, digit1 c, digit2 c, digit3 c, digit4 c, digit5 c, digit6 c, digit7 c]" 
  show ?thesis unfolding integer_of_char_def unfolding of_char_def ds_def[symmetric]
    by (induct ds, auto simp: integer_of_nat_def)
qed

lemma of_char_code[code]: "of_char c = of_nat (nat_of_integer (integer_of_char c))"
proof -
  define ds where "ds = [digit0 c, digit1 c, digit2 c, digit3 c, digit4 c, digit5 c, digit6 c, digit7 c]" 
  show ?thesis 
    unfolding integer_of_char_of_nat nat_of_integer_integer_of_nat
    unfolding of_char_def ds_def[symmetric]
    by (induct ds, auto simp: integer_of_nat_def)
qed

lemma less_char_code[code]: "less c d = (integer_of_char c < integer_of_char d)" 
  unfolding less_char_def unfolding integer_of_char_of_nat integer_of_nat_def by auto

lemma less_eq_char_code[code]: "less_eq c d = (integer_of_char c \<le> integer_of_char d)" 
  unfolding less_eq_char_def unfolding integer_of_char_of_nat integer_of_nat_def by auto

lemma equal_char_code[code]: "equal_class.equal c d = (integer_of_char c = integer_of_char d)" 
  unfolding equal_char_def integer_of_char_of_nat integer_of_nat_def by simp

lemma char_comparison[code_unfold]: "((of_char c :: nat) < of_char d) \<longleftrightarrow> integer_of_char c < integer_of_char d" 
  unfolding integer_of_char_of_nat integer_of_nat_def by simp

(* next we provide a separate implementation for all 256 characters *)
context
begin
  qualified lift_definition char_0 :: char_integer is 0 by auto
  qualified lift_definition char_1 :: char_integer is 1 by auto
  qualified lift_definition char_2 :: char_integer is 2 by auto
  qualified lift_definition char_3 :: char_integer is 3 by auto
  qualified lift_definition char_4 :: char_integer is 4 by auto
  qualified lift_definition char_5 :: char_integer is 5 by auto
  qualified lift_definition char_6 :: char_integer is 6 by auto
  qualified lift_definition char_7 :: char_integer is 7 by auto
  qualified lift_definition char_8 :: char_integer is 8 by auto
  qualified lift_definition char_9 :: char_integer is 9 by auto
  qualified lift_definition char_a :: char_integer is 10 by auto
  qualified lift_definition char_b :: char_integer is 11 by auto
  qualified lift_definition char_c :: char_integer is 12 by auto
  qualified lift_definition char_d :: char_integer is 13 by auto
  qualified lift_definition char_e :: char_integer is 14 by auto
  qualified lift_definition char_f :: char_integer is 15 by auto
  qualified lift_definition char_10 :: char_integer is 16 by auto
  qualified lift_definition char_11 :: char_integer is 17 by auto
  qualified lift_definition char_12 :: char_integer is 18 by auto
  qualified lift_definition char_13 :: char_integer is 19 by auto
  qualified lift_definition char_14 :: char_integer is 20 by auto
  qualified lift_definition char_15 :: char_integer is 21 by auto
  qualified lift_definition char_16 :: char_integer is 22 by auto
  qualified lift_definition char_17 :: char_integer is 23 by auto
  qualified lift_definition char_18 :: char_integer is 24 by auto
  qualified lift_definition char_19 :: char_integer is 25 by auto
  qualified lift_definition char_1a :: char_integer is 26 by auto
  qualified lift_definition char_1b :: char_integer is 27 by auto
  qualified lift_definition char_1c :: char_integer is 28 by auto
  qualified lift_definition char_1d :: char_integer is 29 by auto
  qualified lift_definition char_1e :: char_integer is 30 by auto
  qualified lift_definition char_1f :: char_integer is 31 by auto
  qualified lift_definition char_20 :: char_integer is 32 by auto
  qualified lift_definition char_21 :: char_integer is 33 by auto
  qualified lift_definition char_22 :: char_integer is 34 by auto
  qualified lift_definition char_23 :: char_integer is 35 by auto
  qualified lift_definition char_24 :: char_integer is 36 by auto
  qualified lift_definition char_25 :: char_integer is 37 by auto
  qualified lift_definition char_26 :: char_integer is 38 by auto
  qualified lift_definition char_27 :: char_integer is 39 by auto
  qualified lift_definition char_28 :: char_integer is 40 by auto
  qualified lift_definition char_29 :: char_integer is 41 by auto
  qualified lift_definition char_2a :: char_integer is 42 by auto
  qualified lift_definition char_2b :: char_integer is 43 by auto
  qualified lift_definition char_2c :: char_integer is 44 by auto
  qualified lift_definition char_2d :: char_integer is 45 by auto
  qualified lift_definition char_2e :: char_integer is 46 by auto
  qualified lift_definition char_2f :: char_integer is 47 by auto
  qualified lift_definition char_30 :: char_integer is 48 by auto
  qualified lift_definition char_31 :: char_integer is 49 by auto
  qualified lift_definition char_32 :: char_integer is 50 by auto
  qualified lift_definition char_33 :: char_integer is 51 by auto
  qualified lift_definition char_34 :: char_integer is 52 by auto
  qualified lift_definition char_35 :: char_integer is 53 by auto
  qualified lift_definition char_36 :: char_integer is 54 by auto
  qualified lift_definition char_37 :: char_integer is 55 by auto
  qualified lift_definition char_38 :: char_integer is 56 by auto
  qualified lift_definition char_39 :: char_integer is 57 by auto
  qualified lift_definition char_3a :: char_integer is 58 by auto
  qualified lift_definition char_3b :: char_integer is 59 by auto
  qualified lift_definition char_3c :: char_integer is 60 by auto
  qualified lift_definition char_3d :: char_integer is 61 by auto
  qualified lift_definition char_3e :: char_integer is 62 by auto
  qualified lift_definition char_3f :: char_integer is 63 by auto
  qualified lift_definition char_40 :: char_integer is 64 by auto
  qualified lift_definition char_41 :: char_integer is 65 by auto
  qualified lift_definition char_42 :: char_integer is 66 by auto
  qualified lift_definition char_43 :: char_integer is 67 by auto
  qualified lift_definition char_44 :: char_integer is 68 by auto
  qualified lift_definition char_45 :: char_integer is 69 by auto
  qualified lift_definition char_46 :: char_integer is 70 by auto
  qualified lift_definition char_47 :: char_integer is 71 by auto
  qualified lift_definition char_48 :: char_integer is 72 by auto
  qualified lift_definition char_49 :: char_integer is 73 by auto
  qualified lift_definition char_4a :: char_integer is 74 by auto
  qualified lift_definition char_4b :: char_integer is 75 by auto
  qualified lift_definition char_4c :: char_integer is 76 by auto
  qualified lift_definition char_4d :: char_integer is 77 by auto
  qualified lift_definition char_4e :: char_integer is 78 by auto
  qualified lift_definition char_4f :: char_integer is 79 by auto
  qualified lift_definition char_50 :: char_integer is 80 by auto
  qualified lift_definition char_51 :: char_integer is 81 by auto
  qualified lift_definition char_52 :: char_integer is 82 by auto
  qualified lift_definition char_53 :: char_integer is 83 by auto
  qualified lift_definition char_54 :: char_integer is 84 by auto
  qualified lift_definition char_55 :: char_integer is 85 by auto
  qualified lift_definition char_56 :: char_integer is 86 by auto
  qualified lift_definition char_57 :: char_integer is 87 by auto
  qualified lift_definition char_58 :: char_integer is 88 by auto
  qualified lift_definition char_59 :: char_integer is 89 by auto
  qualified lift_definition char_5a :: char_integer is 90 by auto
  qualified lift_definition char_5b :: char_integer is 91 by auto
  qualified lift_definition char_5c :: char_integer is 92 by auto
  qualified lift_definition char_5d :: char_integer is 93 by auto
  qualified lift_definition char_5e :: char_integer is 94 by auto
  qualified lift_definition char_5f :: char_integer is 95 by auto
  qualified lift_definition char_60 :: char_integer is 96 by auto
  qualified lift_definition char_61 :: char_integer is 97 by auto
  qualified lift_definition char_62 :: char_integer is 98 by auto
  qualified lift_definition char_63 :: char_integer is 99 by auto
  qualified lift_definition char_64 :: char_integer is 100 by auto
  qualified lift_definition char_65 :: char_integer is 101 by auto
  qualified lift_definition char_66 :: char_integer is 102 by auto
  qualified lift_definition char_67 :: char_integer is 103 by auto
  qualified lift_definition char_68 :: char_integer is 104 by auto
  qualified lift_definition char_69 :: char_integer is 105 by auto
  qualified lift_definition char_6a :: char_integer is 106 by auto
  qualified lift_definition char_6b :: char_integer is 107 by auto
  qualified lift_definition char_6c :: char_integer is 108 by auto
  qualified lift_definition char_6d :: char_integer is 109 by auto
  qualified lift_definition char_6e :: char_integer is 110 by auto
  qualified lift_definition char_6f :: char_integer is 111 by auto
  qualified lift_definition char_70 :: char_integer is 112 by auto
  qualified lift_definition char_71 :: char_integer is 113 by auto
  qualified lift_definition char_72 :: char_integer is 114 by auto
  qualified lift_definition char_73 :: char_integer is 115 by auto
  qualified lift_definition char_74 :: char_integer is 116 by auto
  qualified lift_definition char_75 :: char_integer is 117 by auto
  qualified lift_definition char_76 :: char_integer is 118 by auto
  qualified lift_definition char_77 :: char_integer is 119 by auto
  qualified lift_definition char_78 :: char_integer is 120 by auto
  qualified lift_definition char_79 :: char_integer is 121 by auto
  qualified lift_definition char_7a :: char_integer is 122 by auto
  qualified lift_definition char_7b :: char_integer is 123 by auto
  qualified lift_definition char_7c :: char_integer is 124 by auto
  qualified lift_definition char_7d :: char_integer is 125 by auto
  qualified lift_definition char_7e :: char_integer is 126 by auto
  qualified lift_definition char_7f :: char_integer is 127 by auto
  qualified lift_definition char_80 :: char_integer is 128 by auto
  qualified lift_definition char_81 :: char_integer is 129 by auto
  qualified lift_definition char_82 :: char_integer is 130 by auto
  qualified lift_definition char_83 :: char_integer is 131 by auto
  qualified lift_definition char_84 :: char_integer is 132 by auto
  qualified lift_definition char_85 :: char_integer is 133 by auto
  qualified lift_definition char_86 :: char_integer is 134 by auto
  qualified lift_definition char_87 :: char_integer is 135 by auto
  qualified lift_definition char_88 :: char_integer is 136 by auto
  qualified lift_definition char_89 :: char_integer is 137 by auto
  qualified lift_definition char_8a :: char_integer is 138 by auto
  qualified lift_definition char_8b :: char_integer is 139 by auto
  qualified lift_definition char_8c :: char_integer is 140 by auto
  qualified lift_definition char_8d :: char_integer is 141 by auto
  qualified lift_definition char_8e :: char_integer is 142 by auto
  qualified lift_definition char_8f :: char_integer is 143 by auto
  qualified lift_definition char_90 :: char_integer is 144 by auto
  qualified lift_definition char_91 :: char_integer is 145 by auto
  qualified lift_definition char_92 :: char_integer is 146 by auto
  qualified lift_definition char_93 :: char_integer is 147 by auto
  qualified lift_definition char_94 :: char_integer is 148 by auto
  qualified lift_definition char_95 :: char_integer is 149 by auto
  qualified lift_definition char_96 :: char_integer is 150 by auto
  qualified lift_definition char_97 :: char_integer is 151 by auto
  qualified lift_definition char_98 :: char_integer is 152 by auto
  qualified lift_definition char_99 :: char_integer is 153 by auto
  qualified lift_definition char_9a :: char_integer is 154 by auto
  qualified lift_definition char_9b :: char_integer is 155 by auto
  qualified lift_definition char_9c :: char_integer is 156 by auto
  qualified lift_definition char_9d :: char_integer is 157 by auto
  qualified lift_definition char_9e :: char_integer is 158 by auto
  qualified lift_definition char_9f :: char_integer is 159 by auto
  qualified lift_definition char_a0 :: char_integer is 160 by auto
  qualified lift_definition char_a1 :: char_integer is 161 by auto
  qualified lift_definition char_a2 :: char_integer is 162 by auto
  qualified lift_definition char_a3 :: char_integer is 163 by auto
  qualified lift_definition char_a4 :: char_integer is 164 by auto
  qualified lift_definition char_a5 :: char_integer is 165 by auto
  qualified lift_definition char_a6 :: char_integer is 166 by auto
  qualified lift_definition char_a7 :: char_integer is 167 by auto
  qualified lift_definition char_a8 :: char_integer is 168 by auto
  qualified lift_definition char_a9 :: char_integer is 169 by auto
  qualified lift_definition char_aa :: char_integer is 170 by auto
  qualified lift_definition char_ab :: char_integer is 171 by auto
  qualified lift_definition char_ac :: char_integer is 172 by auto
  qualified lift_definition char_ad :: char_integer is 173 by auto
  qualified lift_definition char_ae :: char_integer is 174 by auto
  qualified lift_definition char_af :: char_integer is 175 by auto
  qualified lift_definition char_b0 :: char_integer is 176 by auto
  qualified lift_definition char_b1 :: char_integer is 177 by auto
  qualified lift_definition char_b2 :: char_integer is 178 by auto
  qualified lift_definition char_b3 :: char_integer is 179 by auto
  qualified lift_definition char_b4 :: char_integer is 180 by auto
  qualified lift_definition char_b5 :: char_integer is 181 by auto
  qualified lift_definition char_b6 :: char_integer is 182 by auto
  qualified lift_definition char_b7 :: char_integer is 183 by auto
  qualified lift_definition char_b8 :: char_integer is 184 by auto
  qualified lift_definition char_b9 :: char_integer is 185 by auto
  qualified lift_definition char_ba :: char_integer is 186 by auto
  qualified lift_definition char_bb :: char_integer is 187 by auto
  qualified lift_definition char_bc :: char_integer is 188 by auto
  qualified lift_definition char_bd :: char_integer is 189 by auto
  qualified lift_definition char_be :: char_integer is 190 by auto
  qualified lift_definition char_bf :: char_integer is 191 by auto
  qualified lift_definition char_c0 :: char_integer is 192 by auto
  qualified lift_definition char_c1 :: char_integer is 193 by auto
  qualified lift_definition char_c2 :: char_integer is 194 by auto
  qualified lift_definition char_c3 :: char_integer is 195 by auto
  qualified lift_definition char_c4 :: char_integer is 196 by auto
  qualified lift_definition char_c5 :: char_integer is 197 by auto
  qualified lift_definition char_c6 :: char_integer is 198 by auto
  qualified lift_definition char_c7 :: char_integer is 199 by auto
  qualified lift_definition char_c8 :: char_integer is 200 by auto
  qualified lift_definition char_c9 :: char_integer is 201 by auto
  qualified lift_definition char_ca :: char_integer is 202 by auto
  qualified lift_definition char_cb :: char_integer is 203 by auto
  qualified lift_definition char_cc :: char_integer is 204 by auto
  qualified lift_definition char_cd :: char_integer is 205 by auto
  qualified lift_definition char_ce :: char_integer is 206 by auto
  qualified lift_definition char_cf :: char_integer is 207 by auto
  qualified lift_definition char_d0 :: char_integer is 208 by auto
  qualified lift_definition char_d1 :: char_integer is 209 by auto
  qualified lift_definition char_d2 :: char_integer is 210 by auto
  qualified lift_definition char_d3 :: char_integer is 211 by auto
  qualified lift_definition char_d4 :: char_integer is 212 by auto
  qualified lift_definition char_d5 :: char_integer is 213 by auto
  qualified lift_definition char_d6 :: char_integer is 214 by auto
  qualified lift_definition char_d7 :: char_integer is 215 by auto
  qualified lift_definition char_d8 :: char_integer is 216 by auto
  qualified lift_definition char_d9 :: char_integer is 217 by auto
  qualified lift_definition char_da :: char_integer is 218 by auto
  qualified lift_definition char_db :: char_integer is 219 by auto
  qualified lift_definition char_dc :: char_integer is 220 by auto
  qualified lift_definition char_dd :: char_integer is 221 by auto
  qualified lift_definition char_de :: char_integer is 222 by auto
  qualified lift_definition char_df :: char_integer is 223 by auto
  qualified lift_definition char_e0 :: char_integer is 224 by auto
  qualified lift_definition char_e1 :: char_integer is 225 by auto
  qualified lift_definition char_e2 :: char_integer is 226 by auto
  qualified lift_definition char_e3 :: char_integer is 227 by auto
  qualified lift_definition char_e4 :: char_integer is 228 by auto
  qualified lift_definition char_e5 :: char_integer is 229 by auto
  qualified lift_definition char_e6 :: char_integer is 230 by auto
  qualified lift_definition char_e7 :: char_integer is 231 by auto
  qualified lift_definition char_e8 :: char_integer is 232 by auto
  qualified lift_definition char_e9 :: char_integer is 233 by auto
  qualified lift_definition char_ea :: char_integer is 234 by auto
  qualified lift_definition char_eb :: char_integer is 235 by auto
  qualified lift_definition char_ec :: char_integer is 236 by auto
  qualified lift_definition char_ed :: char_integer is 237 by auto
  qualified lift_definition char_ee :: char_integer is 238 by auto
  qualified lift_definition char_ef :: char_integer is 239 by auto
  qualified lift_definition char_f0 :: char_integer is 240 by auto
  qualified lift_definition char_f1 :: char_integer is 241 by auto
  qualified lift_definition char_f2 :: char_integer is 242 by auto
  qualified lift_definition char_f3 :: char_integer is 243 by auto
  qualified lift_definition char_f4 :: char_integer is 244 by auto
  qualified lift_definition char_f5 :: char_integer is 245 by auto
  qualified lift_definition char_f6 :: char_integer is 246 by auto
  qualified lift_definition char_f7 :: char_integer is 247 by auto
  qualified lift_definition char_f8 :: char_integer is 248 by auto
  qualified lift_definition char_f9 :: char_integer is 249 by auto
  qualified lift_definition char_fa :: char_integer is 250 by auto
  qualified lift_definition char_fb :: char_integer is 251 by auto
  qualified lift_definition char_fc :: char_integer is 252 by auto
  qualified lift_definition char_fd :: char_integer is 253 by auto
  qualified lift_definition char_fe :: char_integer is 254 by auto
  qualified lift_definition char_ff :: char_integer is 255 by auto


lemma char_256_unfold [code_unfold]: 
  "CHR 0x00 = char_of_ci char_0" 
  "CHR 0x01 = char_of_ci char_1" 
  "CHR 0x02 = char_of_ci char_2" 
  "CHR 0x03 = char_of_ci char_3" 
  "CHR 0x04 = char_of_ci char_4" 
  "CHR 0x05 = char_of_ci char_5" 
  "CHR 0x06 = char_of_ci char_6" 
  "CHR 0x07 = char_of_ci char_7" 
  "CHR 0x08 = char_of_ci char_8" 
  "CHR 0x09 = char_of_ci char_9" 
  "CHR 0x0a = char_of_ci char_a" 
  "CHR 0x0b = char_of_ci char_b" 
  "CHR 0x0c = char_of_ci char_c" 
  "CHR 0x0d = char_of_ci char_d" 
  "CHR 0x0e = char_of_ci char_e" 
  "CHR 0x0f = char_of_ci char_f" 
  "CHR 0x10 = char_of_ci char_10" 
  "CHR 0x11 = char_of_ci char_11" 
  "CHR 0x12 = char_of_ci char_12" 
  "CHR 0x13 = char_of_ci char_13" 
  "CHR 0x14 = char_of_ci char_14" 
  "CHR 0x15 = char_of_ci char_15" 
  "CHR 0x16 = char_of_ci char_16" 
  "CHR 0x17 = char_of_ci char_17" 
  "CHR 0x18 = char_of_ci char_18" 
  "CHR 0x19 = char_of_ci char_19" 
  "CHR 0x1a = char_of_ci char_1a" 
  "CHR 0x1b = char_of_ci char_1b" 
  "CHR 0x1c = char_of_ci char_1c" 
  "CHR 0x1d = char_of_ci char_1d" 
  "CHR 0x1e = char_of_ci char_1e" 
  "CHR 0x1f = char_of_ci char_1f" 
  "CHR 0x20 = char_of_ci char_20" 
  "CHR 0x21 = char_of_ci char_21" 
  "CHR 0x22 = char_of_ci char_22" 
  "CHR 0x23 = char_of_ci char_23" 
  "CHR 0x24 = char_of_ci char_24" 
  "CHR 0x25 = char_of_ci char_25" 
  "CHR 0x26 = char_of_ci char_26" 
  "CHR 0x27 = char_of_ci char_27" 
  "CHR 0x28 = char_of_ci char_28" 
  "CHR 0x29 = char_of_ci char_29" 
  "CHR 0x2a = char_of_ci char_2a" 
  "CHR 0x2b = char_of_ci char_2b" 
  "CHR 0x2c = char_of_ci char_2c" 
  "CHR 0x2d = char_of_ci char_2d" 
  "CHR 0x2e = char_of_ci char_2e" 
  "CHR 0x2f = char_of_ci char_2f" 
  "CHR 0x30 = char_of_ci char_30" 
  "CHR 0x31 = char_of_ci char_31" 
  "CHR 0x32 = char_of_ci char_32" 
  "CHR 0x33 = char_of_ci char_33" 
  "CHR 0x34 = char_of_ci char_34" 
  "CHR 0x35 = char_of_ci char_35" 
  "CHR 0x36 = char_of_ci char_36" 
  "CHR 0x37 = char_of_ci char_37" 
  "CHR 0x38 = char_of_ci char_38" 
  "CHR 0x39 = char_of_ci char_39" 
  "CHR 0x3a = char_of_ci char_3a" 
  "CHR 0x3b = char_of_ci char_3b" 
  "CHR 0x3c = char_of_ci char_3c" 
  "CHR 0x3d = char_of_ci char_3d" 
  "CHR 0x3e = char_of_ci char_3e" 
  "CHR 0x3f = char_of_ci char_3f" 
  "CHR 0x40 = char_of_ci char_40" 
  "CHR 0x41 = char_of_ci char_41" 
  "CHR 0x42 = char_of_ci char_42" 
  "CHR 0x43 = char_of_ci char_43" 
  "CHR 0x44 = char_of_ci char_44" 
  "CHR 0x45 = char_of_ci char_45" 
  "CHR 0x46 = char_of_ci char_46" 
  "CHR 0x47 = char_of_ci char_47" 
  "CHR 0x48 = char_of_ci char_48" 
  "CHR 0x49 = char_of_ci char_49" 
  "CHR 0x4a = char_of_ci char_4a" 
  "CHR 0x4b = char_of_ci char_4b" 
  "CHR 0x4c = char_of_ci char_4c" 
  "CHR 0x4d = char_of_ci char_4d" 
  "CHR 0x4e = char_of_ci char_4e" 
  "CHR 0x4f = char_of_ci char_4f" 
  "CHR 0x50 = char_of_ci char_50" 
  "CHR 0x51 = char_of_ci char_51" 
  "CHR 0x52 = char_of_ci char_52" 
  "CHR 0x53 = char_of_ci char_53" 
  "CHR 0x54 = char_of_ci char_54" 
  "CHR 0x55 = char_of_ci char_55" 
  "CHR 0x56 = char_of_ci char_56" 
  "CHR 0x57 = char_of_ci char_57" 
  "CHR 0x58 = char_of_ci char_58" 
  "CHR 0x59 = char_of_ci char_59" 
  "CHR 0x5a = char_of_ci char_5a" 
  "CHR 0x5b = char_of_ci char_5b" 
  "CHR 0x5c = char_of_ci char_5c" 
  "CHR 0x5d = char_of_ci char_5d" 
  "CHR 0x5e = char_of_ci char_5e" 
  "CHR 0x5f = char_of_ci char_5f" 
  "CHR 0x60 = char_of_ci char_60" 
  "CHR 0x61 = char_of_ci char_61" 
  "CHR 0x62 = char_of_ci char_62" 
  "CHR 0x63 = char_of_ci char_63" 
  "CHR 0x64 = char_of_ci char_64" 
  "CHR 0x65 = char_of_ci char_65" 
  "CHR 0x66 = char_of_ci char_66" 
  "CHR 0x67 = char_of_ci char_67" 
  "CHR 0x68 = char_of_ci char_68" 
  "CHR 0x69 = char_of_ci char_69" 
  "CHR 0x6a = char_of_ci char_6a" 
  "CHR 0x6b = char_of_ci char_6b" 
  "CHR 0x6c = char_of_ci char_6c" 
  "CHR 0x6d = char_of_ci char_6d" 
  "CHR 0x6e = char_of_ci char_6e" 
  "CHR 0x6f = char_of_ci char_6f" 
  "CHR 0x70 = char_of_ci char_70" 
  "CHR 0x71 = char_of_ci char_71" 
  "CHR 0x72 = char_of_ci char_72" 
  "CHR 0x73 = char_of_ci char_73" 
  "CHR 0x74 = char_of_ci char_74" 
  "CHR 0x75 = char_of_ci char_75" 
  "CHR 0x76 = char_of_ci char_76" 
  "CHR 0x77 = char_of_ci char_77" 
  "CHR 0x78 = char_of_ci char_78" 
  "CHR 0x79 = char_of_ci char_79" 
  "CHR 0x7a = char_of_ci char_7a" 
  "CHR 0x7b = char_of_ci char_7b" 
  "CHR 0x7c = char_of_ci char_7c" 
  "CHR 0x7d = char_of_ci char_7d" 
  "CHR 0x7e = char_of_ci char_7e" 
  "CHR 0x7f = char_of_ci char_7f" 
  "CHR 0x80 = char_of_ci char_80" 
  "CHR 0x81 = char_of_ci char_81" 
  "CHR 0x82 = char_of_ci char_82" 
  "CHR 0x83 = char_of_ci char_83" 
  "CHR 0x84 = char_of_ci char_84" 
  "CHR 0x85 = char_of_ci char_85" 
  "CHR 0x86 = char_of_ci char_86" 
  "CHR 0x87 = char_of_ci char_87" 
  "CHR 0x88 = char_of_ci char_88" 
  "CHR 0x89 = char_of_ci char_89" 
  "CHR 0x8a = char_of_ci char_8a" 
  "CHR 0x8b = char_of_ci char_8b" 
  "CHR 0x8c = char_of_ci char_8c" 
  "CHR 0x8d = char_of_ci char_8d" 
  "CHR 0x8e = char_of_ci char_8e" 
  "CHR 0x8f = char_of_ci char_8f" 
  "CHR 0x90 = char_of_ci char_90" 
  "CHR 0x91 = char_of_ci char_91" 
  "CHR 0x92 = char_of_ci char_92" 
  "CHR 0x93 = char_of_ci char_93" 
  "CHR 0x94 = char_of_ci char_94" 
  "CHR 0x95 = char_of_ci char_95" 
  "CHR 0x96 = char_of_ci char_96" 
  "CHR 0x97 = char_of_ci char_97" 
  "CHR 0x98 = char_of_ci char_98" 
  "CHR 0x99 = char_of_ci char_99" 
  "CHR 0x9a = char_of_ci char_9a" 
  "CHR 0x9b = char_of_ci char_9b" 
  "CHR 0x9c = char_of_ci char_9c" 
  "CHR 0x9d = char_of_ci char_9d" 
  "CHR 0x9e = char_of_ci char_9e" 
  "CHR 0x9f = char_of_ci char_9f" 
  "CHR 0xa0 = char_of_ci char_a0" 
  "CHR 0xa1 = char_of_ci char_a1" 
  "CHR 0xa2 = char_of_ci char_a2" 
  "CHR 0xa3 = char_of_ci char_a3" 
  "CHR 0xa4 = char_of_ci char_a4" 
  "CHR 0xa5 = char_of_ci char_a5" 
  "CHR 0xa6 = char_of_ci char_a6" 
  "CHR 0xa7 = char_of_ci char_a7" 
  "CHR 0xa8 = char_of_ci char_a8" 
  "CHR 0xa9 = char_of_ci char_a9" 
  "CHR 0xaa = char_of_ci char_aa" 
  "CHR 0xab = char_of_ci char_ab" 
  "CHR 0xac = char_of_ci char_ac" 
  "CHR 0xad = char_of_ci char_ad" 
  "CHR 0xae = char_of_ci char_ae" 
  "CHR 0xaf = char_of_ci char_af" 
  "CHR 0xb0 = char_of_ci char_b0" 
  "CHR 0xb1 = char_of_ci char_b1" 
  "CHR 0xb2 = char_of_ci char_b2" 
  "CHR 0xb3 = char_of_ci char_b3" 
  "CHR 0xb4 = char_of_ci char_b4" 
  "CHR 0xb5 = char_of_ci char_b5" 
  "CHR 0xb6 = char_of_ci char_b6" 
  "CHR 0xb7 = char_of_ci char_b7" 
  "CHR 0xb8 = char_of_ci char_b8" 
  "CHR 0xb9 = char_of_ci char_b9" 
  "CHR 0xba = char_of_ci char_ba" 
  "CHR 0xbb = char_of_ci char_bb" 
  "CHR 0xbc = char_of_ci char_bc" 
  "CHR 0xbd = char_of_ci char_bd" 
  "CHR 0xbe = char_of_ci char_be" 
  "CHR 0xbf = char_of_ci char_bf" 
  "CHR 0xc0 = char_of_ci char_c0" 
  "CHR 0xc1 = char_of_ci char_c1" 
  "CHR 0xc2 = char_of_ci char_c2" 
  "CHR 0xc3 = char_of_ci char_c3" 
  "CHR 0xc4 = char_of_ci char_c4" 
  "CHR 0xc5 = char_of_ci char_c5" 
  "CHR 0xc6 = char_of_ci char_c6" 
  "CHR 0xc7 = char_of_ci char_c7" 
  "CHR 0xc8 = char_of_ci char_c8" 
  "CHR 0xc9 = char_of_ci char_c9" 
  "CHR 0xca = char_of_ci char_ca" 
  "CHR 0xcb = char_of_ci char_cb" 
  "CHR 0xcc = char_of_ci char_cc" 
  "CHR 0xcd = char_of_ci char_cd" 
  "CHR 0xce = char_of_ci char_ce" 
  "CHR 0xcf = char_of_ci char_cf" 
  "CHR 0xd0 = char_of_ci char_d0" 
  "CHR 0xd1 = char_of_ci char_d1" 
  "CHR 0xd2 = char_of_ci char_d2" 
  "CHR 0xd3 = char_of_ci char_d3" 
  "CHR 0xd4 = char_of_ci char_d4" 
  "CHR 0xd5 = char_of_ci char_d5" 
  "CHR 0xd6 = char_of_ci char_d6" 
  "CHR 0xd7 = char_of_ci char_d7" 
  "CHR 0xd8 = char_of_ci char_d8" 
  "CHR 0xd9 = char_of_ci char_d9" 
  "CHR 0xda = char_of_ci char_da" 
  "CHR 0xdb = char_of_ci char_db" 
  "CHR 0xdc = char_of_ci char_dc" 
  "CHR 0xdd = char_of_ci char_dd" 
  "CHR 0xde = char_of_ci char_de" 
  "CHR 0xdf = char_of_ci char_df" 
  "CHR 0xe0 = char_of_ci char_e0" 
  "CHR 0xe1 = char_of_ci char_e1" 
  "CHR 0xe2 = char_of_ci char_e2" 
  "CHR 0xe3 = char_of_ci char_e3" 
  "CHR 0xe4 = char_of_ci char_e4" 
  "CHR 0xe5 = char_of_ci char_e5" 
  "CHR 0xe6 = char_of_ci char_e6" 
  "CHR 0xe7 = char_of_ci char_e7" 
  "CHR 0xe8 = char_of_ci char_e8" 
  "CHR 0xe9 = char_of_ci char_e9" 
  "CHR 0xea = char_of_ci char_ea" 
  "CHR 0xeb = char_of_ci char_eb" 
  "CHR 0xec = char_of_ci char_ec" 
  "CHR 0xed = char_of_ci char_ed" 
  "CHR 0xee = char_of_ci char_ee" 
  "CHR 0xef = char_of_ci char_ef" 
  "CHR 0xf0 = char_of_ci char_f0" 
  "CHR 0xf1 = char_of_ci char_f1" 
  "CHR 0xf2 = char_of_ci char_f2" 
  "CHR 0xf3 = char_of_ci char_f3" 
  "CHR 0xf4 = char_of_ci char_f4" 
  "CHR 0xf5 = char_of_ci char_f5" 
  "CHR 0xf6 = char_of_ci char_f6" 
  "CHR 0xf7 = char_of_ci char_f7" 
  "CHR 0xf8 = char_of_ci char_f8" 
  "CHR 0xf9 = char_of_ci char_f9" 
  "CHR 0xfa = char_of_ci char_fa" 
  "CHR 0xfb = char_of_ci char_fb" 
  "CHR 0xfc = char_of_ci char_fc" 
  "CHR 0xfd = char_of_ci char_fd" 
  "CHR 0xfe = char_of_ci char_fe" 
  "CHR 0xff = char_of_ci char_ff" 
  by (transfer, simp add: char_of_integer_def char_of_def)+
end

end